# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse

from . import server
from .config import ModelExplorerConfig
from .consts import DEFAULT_HOST, DEFAULT_PORT

parser = argparse.ArgumentParser(
    prog='model-explorer',
    description='A modern model graph visualizer and debugger',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    'model_paths', nargs='?', help='comma separated model file paths'
)
parser.add_argument('--host', default=DEFAULT_HOST, help='host of the server')
parser.add_argument(
    '--port', default=DEFAULT_PORT, type=int, help='port of the server'
)
parser.add_argument(
    '--node_data_paths', help='comma separated node data file paths'
)
parser.add_argument(
    '--no_open_in_browser',
    action='store_true',
    help="Don't open the web app in browser after server starts",
)
parser.add_argument(
    '--extensions', help='comma separated extension module names'
)
parser.add_argument(
    '--cors_host', help='the host of the Access-Control-Allow-Origin header'
)
parser.add_argument(
    '--skip_health_check',
    action='store_true',
    help='Whether to skip the health check after server starts',
)
parser.add_argument(
    '--reuse_server',
    action='store_true',
    help='Whether to reuse the currently running server',
)
parser.add_argument(
    '--reuse_server_host',
    default=DEFAULT_HOST,
    help='The host of the server to reuse',
)
parser.add_argument(
    '--reuse_server_port',
    default=-1,
    type=int,
    help=(
        'The port of the server to reuse. If unspecified, '
        'it will try to find a running server from port 8080 '
        'to 8099'
    ),
)
args = parser.parse_args()


def main():
  """Entry point for the command line version of model explorer."""

  model_paths: list[str] = []
  if args.model_paths is not None and args.model_paths != '':
    model_paths = [x.strip() for x in args.model_paths.split(',')]

  node_data_paths: list[str] = []
  if args.node_data_paths is not None and args.node_data_paths != '':
    node_data_paths = [x.strip() for x in args.node_data_paths.split(',')]

  extensions: list[str] = []
  if args.extensions is not None:
    extensions = [x.strip() for x in args.extensions.split(',')]

  # Construct config.
  config = ModelExplorerConfig()
  for model_path in model_paths:
    config.add_model_from_path(model_path)
  for node_data_path in node_data_paths:
    config.add_node_data_from_path(node_data_path)
  if args.reuse_server:
    config.set_reuse_server(
        server_host=args.reuse_server_host, server_port=args.reuse_server_port
    )

  server.start(
      host=args.host,
      port=args.port,
      config=config,
      extensions=extensions,
      cors_host=args.cors_host,
      no_open_in_browser=args.no_open_in_browser,
      skip_health_check=args.skip_health_check,
  )



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from .extension_matadata import ExtensionMetadata
from .types import ExtensionClassType


class RegisteredExtension(object):
  """A registered extension in extension manager."""

  def __init__(
      self,
      metadata: ExtensionMetadata,
      type: str,
      ext_class: ExtensionClassType,
  ):
    self.metadata = metadata
    self.type = type
    self.ext_class = ext_class



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass, field
from typing import Dict, Any

from .extension_base import Extension
from .extension_matadata import ExtensionMetadata
from .types import ModelExplorerGraphs


@dataclass
class AdapterMetadata(ExtensionMetadata):
  """Metadata for a adapter extension."""

  # Supported model file extensions.
  fileExts: list[str] = field(default_factory=list)


class Adapter(Extension):
  """The base class of a adapter extension."""

  # Base class needs to override this class variable to provide its metadata.
  metadata = AdapterMetadata()

  def __init__(self):
    Extension.__init__(self)

  def convert(self, model_path: str, settings: Dict) -> ModelExplorerGraphs:
    """Converts the given model with the given settings.

    Args:
      model_path: the absolute path of the model.
      settings: settings set from the frontend (TBD).

    Returns:
      A list of graphs, or a list of graph collections.

      {
        graphs,
        graphCollections,
      }
    """
    return {}



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from .types import AdapterCommand, ModelExplorerGraphs

from .registered_extension import RegisteredExtension
from .utils import get_instance_method


class AdapterRunner:
  """A runner to run adapter extension."""

  def run_adapter(
      self, extension: RegisteredExtension, cmd: AdapterCommand
  ) -> ModelExplorerGraphs:
    # Get extension class and create an instance.
    extension_class = extension.ext_class
    instance = extension_class()

    # Get the method by the cmdId.
    fn = get_instance_method(instance, cmd['cmdId'])
    if fn is None:
      raise Exception(
          f'Method {cmd["cmdId"]} not implemented in the adapter extension'
          f' "{extension.metadata.name}"'
      )
    modelPath = os.path.expanduser(cmd['modelPath'])
    return fn(modelPath, cmd['settings'])



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import logging
import os
import platform
import queue
import socket
import sys
import tempfile
import threading
import time
import traceback
import webbrowser
from importlib import metadata
from time import sleep
from typing import Any, Union

import portpicker
import requests
from flask import Flask, Response, make_response, request, send_from_directory
from IPython import display
from packaging.version import parse
from termcolor import colored, cprint

from .config import ModelExplorerConfig
from .consts import (
    DEFAULT_COLAB_HEIGHT,
    DEFAULT_HOST,
    DEFAULT_PORT,
    PACKAGE_NAME,
)
from .extension_manager import ExtensionManager
from .server_directive_dispatcher import ServerDirectiveDispatcher
from .server_director import ServerDirector
from .utils import convert_adapter_response

server_directive_dispatcher = ServerDirectiveDispatcher()


def _make_json_response(obj):
  body = json.dumps(obj)
  resp = make_response(body)
  resp.headers['Content-Type'] = 'application/json'
  return resp


def _get_latest_version_from_repo(package_json_url: str) -> str:
  req = requests.get(package_json_url)
  version = parse('0')
  if req.status_code == requests.codes.ok:
    j = json.loads(req.text.encode('utf-8'))
    releases = j.get('releases', [])
    for release in releases:
      ver = parse(release)
      if not ver.is_prerelease:
        version = max(version, ver)
  return str(version)


def _get_release_from_github(version: str) -> dict:
  # Get release data through github API.
  req = requests.get(
      f'https://api.github.com/repos/google-ai-edge/model-explorer/releases/tags/model-explorer-v{version}'
  )
  req_json = json.loads(req.text.encode('utf-8'))

  # Construct the search term from platform and cpu architecture for finding
  # asset download url.
  #
  # darwin, linux, etc.
  cur_platform = sys.platform
  # x64_64, arm64, etc.
  cur_mach = platform.machine()
  if cur_mach == 'x86_64':
    cur_mach = 'x64'

  # Find the download url from assets.
  asset_search_term = f'{cur_platform}-{cur_mach}-{version}'
  asset_url = ''
  assets = req_json.get('assets', [])
  for asset in assets:
    browser_download_url = asset.get('browser_download_url', '')
    if asset_search_term in browser_download_url:
      asset_url = browser_download_url
      break

  return {
      'releaseUrl': req_json.get('html_url', ''),
      'desktopAppUrl': f'{asset_url}',
  }


def _print_yellow(x):
  return cprint(x, 'yellow')


def _check_new_version(print_msg=True):
  check_new_version_resp = {
      'version': '',
      'runningVersion': '',
      'releaseUrl': '',
      'desktopAppUrl': '',
  }
  try:
    # Get version from repo.
    repo_version = _get_latest_version_from_repo(
        f'https://pypi.python.org/pypi/{PACKAGE_NAME}/json'
    )

    if repo_version != '0':
      # Compare with the local installed version.
      installed_version = metadata.version(PACKAGE_NAME)
      check_new_version_resp['runningVersion'] = installed_version
      if parse(installed_version) < parse(repo_version):
        check_new_version_resp['version'] = repo_version
        if print_msg:
          _print_yellow(
              f'\n{PACKAGE_NAME} version {repo_version} is available, and you'
              f' are using version {installed_version}.'
          )
          _print_yellow('Consider upgrading via the following command:')
          _print_yellow(f'$ pip install -U {PACKAGE_NAME}')

        # Get the corresponding release data from github.
        github_release = _get_release_from_github(repo_version)
        releaseUrl = github_release['releaseUrl']
        check_new_version_resp['releaseUrl'] = releaseUrl
        check_new_version_resp['desktopAppUrl'] = github_release[
            'desktopAppUrl'
        ]

        if print_msg:
          _print_yellow(f'\nRelease notes: {releaseUrl}')
  except:
    pass
  finally:
    return check_new_version_resp


def _is_port_in_use(host: str, port: int) -> bool:
  try:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
      return s.connect_ex((host, port)) == 0
  except socket.gaierror:
    print(
        f'"{host}" cannot be resolved. Try using IP address directly:'
        ' model-explorer --host=127.0.0.1'
    )
    sys.exit(1)


def _js(script):
  display.display(display.Javascript(script))


def start(
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    config: Union[ModelExplorerConfig, None] = None,
    no_open_in_browser: bool = False,
    extensions: list[str] = [],
    colab_height: int = DEFAULT_COLAB_HEIGHT,
    cors_host: Union[str, None] = None,
    skip_health_check: bool = False,
):
  """Starts the local server that serves the web app.

  Args:
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    config: the object that stores the data to be visualized.
    no_open_in_browser: Don't open the web app in browser after server starts.
    extensions: A list of extension module names. Default to empty.
    colab_height: The height of the embedded iFrame when running in colab.
    cors_host: The value of the Access-Control-Allow-Origin header. The header
        won't be present if it is None.
    skip_health_check: Whether to skip the health check after server starts.
  """

  # Don't start the server if user wants to reuse an existing server.
  if (
      config is not None
      and config.reuse_server_host != ''
      and config.reuse_server_port > 0
  ):
    director = ServerDirector(config=config)
    director.update_config()
    return

  # Check whether it is running in colab.
  colab = 'google.colab' in sys.modules or os.getenv('COLAB_RELEASE_TAG')

  # Check port in non-colab environment.
  if not colab:
    # If default port (8080) is specified, try to find an available port
    # automatically.
    #
    # - First, find a port from 8080 to 8099.
    # - If no available port from that range, pick a random port.
    if port == 8080:
      found_port = False
      for i in range(0, 20):
        port = port + i
        if not _is_port_in_use(host, port):
          found_port = True
          break
      if not found_port:
        port = portpicker.pick_unused_port()
    # If a non-default port is specified, output an error message when port
    # is used.
    else:
      if _is_port_in_use(host, port):
        print(colored(f'port {port} already in use.', 'red'))
        sys.exit(1)

  app = Flask(__name__)

  # Disable logging from werkzeug.
  #
  # Without this, flask will show a warning about using dev server (which is OK
  # in our usecase).
  logging.getLogger('werkzeug').disabled = True

  # Disable startup messages.
  cli: Any = sys.modules['flask.cli']
  cli.show_server_banner = lambda *x: None

  # Print a info message when used in colab.
  if colab:
    print('ℹ️ Please re-run the cell in each new session')
    print()

  # Load extensions.
  print('Loading extensions...')
  extension_manager = ExtensionManager(extensions)
  extension_manager.load_extensions()
  extension_metadata_list = extension_manager.get_extensions_metadata()
  num_extensions = len(extension_metadata_list)
  print(
      f'Loaded {num_extensions} extension{"" if num_extensions == 1 else "s"}:'
  )
  for extension in extension_metadata_list:
    print(f' - {extension["name"]}')

  @app.route('/api/v1/check_new_version')
  def check_new_version():
    """Checks new version."""
    return _make_json_response(_check_new_version(False))

  @app.route('/api/v1/get_extensions')
  def get_extensions():
    """Loads all adapter extensions."""
    return _make_json_response(extension_manager.get_extensions_metadata())

  # Note: using "/api/..." for POST requests is not allowed when running in
  # colab.
  @app.route('/apipost/v1/upload', methods=['POST'])
  def upload_file():
    f = request.files['file']
    file_name = f.filename if f.filename is not None else 'no_name'
    tmp_dir = tempfile.mkdtemp()
    file_path = os.path.join(tmp_dir, file_name)
    f.save(file_path)
    return _make_json_response({'path': file_path})

  # Note: using "/api/..." for POST requests is not allowed when running in
  # colab.
  @app.route('/apipost/v1/send_command', methods=['POST'])
  def send_command():
    try:
      resp = extension_manager.run_cmd(request.json)
      return _make_json_response(resp)
    except Exception as err:
      traceback.print_exc()
      return _make_json_response({'error': f'{type(err).__name__}: {str(err)}'})
    finally:
      extension_manager.cleanup(request.json)

  @app.route('/api/v1/load_graphs_json')
  def load_graphs_json():
    if config is None:
      return {}
    graph_index_str = request.args.get('graph_index')
    if graph_index_str is None:
      return {}
    graph_index = int(graph_index_str)
    return _make_json_response(
        convert_adapter_response(config.get_model_explorer_graphs(graph_index))
    )

  @app.route('/api/v1/load_node_data')
  def load_node_data():
    if config is None:
      return {}
    node_data_index_str = request.args.get('node_data_index')
    if node_data_index_str is None:
      return {}
    node_data_index = int(node_data_index_str)
    node_data = config.get_node_data(node_data_index)
    if isinstance(node_data, str):
      json_str = node_data
    else:
      json_str = node_data.to_json_string()
    return _make_json_response({'content': json_str})

  @app.route('/api/v1/read_text_file')
  def read_text_file():
    path = request.args.get('path')
    if path is None:
      return _make_json_response({'error': 'no file path provided'})
    path = os.path.expanduser(path)

    try:
      with open(path, 'r') as file:
        content = file.read()
      return _make_json_response({'content': content})
    except Exception as err:
      return _make_json_response({'error': str(err)})

  @app.route('/check_health')
  def check_health():
    """Serves check_health request."""
    return 'model_explorer_ok'

  @app.route('/apipost/v1/update_config', methods=['POST'])
  def update_config():
    # TODO(do not submit): Update confnig.
    config_data = request.json
    if config and config_data:
      config.set_transferrable_data(config_data)

      # Ask UI to refresh page with the new url.
      server_directive_dispatcher.broadcast(
          json.dumps({
              'name': 'refreshPage',
              'url': f'/?data={config.to_url_param_value()}',
          })
      )

    return ''

  @app.route('/apistream/server_director')
  def server_director_stream():
    def stream():
      directive_queue = server_directive_dispatcher.listen()
      try:
        while True:
          # Try to get a new message.
          try:
            msg = directive_queue.get(block=False)
            yield f'data: {msg}\n\n'
          except queue.Empty:
            # Ignore if there is no new messages.
            pass

          # Keep the connection alive.
          yield ': heartbeat\n\n'
          time.sleep(1)
      except:
        # The client closes the connection (i.e. close the browser tab)
        server_directive_dispatcher.remove_listener(directive_queue)

    headers = {
        'Access-Control-Allow-Origin': '*',
        'Cache-Control': 'no-cache',
    }
    return Response(
        stream(),
        headers=headers,
        content_type='text/event-stream',
        mimetype='text/event-stream',
    )

  @app.route('/')
  def send_index_html():
    """Serves index.html."""
    return send_from_directory('web_app', 'index.html')

  @app.route('/<path:path>')
  def send_static(path):
    """Serves static files."""
    return send_from_directory('web_app', path)

  @app.after_request
  def add_header(response):
    """Adds headers to all responses."""
    # Don't cache any files.
    response.headers['Cache-Control'] = 'public, max-age=0'
    if cors_host is not None:
      response.headers['Access-Control-Allow-Origin'] = cors_host
      response.headers['Access-Control-Allow-Credentials'] = 'true'
      response.headers['Access-Control-Allow-Headers'] = 'Content-Type'
    return response

  def start_server():
    """Starts the server in non-colab environment."""
    app_thread = threading.Thread(target=lambda: app.run(host=host, port=port))
    app_thread.daemon = True
    app_thread.start()

    # Wait for server to start
    if not skip_health_check:
      while True:
        try:
          response = requests.get(f'http://{host}:{port}/')
        except:
          continue

        if response.status_code == 200:
          break

    # Server is ready.
    server_address = f'http://{host}:{port}'
    url_params: list[str] = []
    if config is not None and config.has_data_to_encode_in_url():
      url_params.append(f'data={config.to_url_param_value()}')

    if len(url_params) > 0:
      server_address = f'{server_address}/?{"&".join(url_params)}'
    print(
        f'\nStarting Model Explorer server at:\n{server_address}\n\nPress'
        ' Ctrl+C to stop.'
    )
    if not no_open_in_browser:
      webbrowser.open_new_tab(f'{server_address}')

    # Check installed version vs published version.
    threading.Thread(target=lambda: _check_new_version()).start()

    try:
      while app_thread.is_alive():
        sleep(1)
    except KeyboardInterrupt:
      print('Stopping server...')
      pass

  def embed_in_colab():
    """Embeds the UI in a colab cell."""
    # Disable scrollbar in iFrame's parent div.
    _js('google.colab.output.setIframeHeight(0, true, {maxHeight: 5000})')

    colab_port = portpicker.pick_unused_port()
    shell = """
      (async () => {
          let url = await google.colab.kernel.proxyPort(%PORT%, {"cache": true});
          if ('%HOSTED_RUNTIME%' !== 'True') {
            url = 'http://localhost:' + %PORT% + '/';
          }
          url += '?show_open_in_new_tab=1';
          if ('%DATA_PARAM%' !== '') {
            url += '&%DATA_PARAM%';
          }
          const iframe = document.createElement('iframe');
          iframe.src = url;
          iframe.setAttribute('width', '100%');
          iframe.setAttribute('height', '%HEIGHT%');
          iframe.setAttribute('frameborder', 0);
          iframe.setAttribute('style', 'border: 1px solid #ccc; box-sizing: border-box; margin-top: 12px;');
          document.body.appendChild(iframe);
        })();
    """
    data_param = ''
    if config is not None and config.has_data_to_encode_in_url():
      data_param = f'data={config.to_url_param_value()}'
    replacements = [
        ('%PORT%', f'{colab_port}'),
        ('%HOSTED_RUNTIME%', f'{colab}'),
        ('%HEIGHT%', f'{colab_height}'),
        ('%DATA_PARAM%', f'{data_param}'),
    ]
    for k, v in replacements:
      shell = shell.replace(k, v)

    threading.Thread(
        target=app.run, kwargs={'host': '::', 'port': colab_port}
    ).start()

    _js(shell)

  if colab:
    embed_in_colab()
  else:
    start_server()



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

PACKAGE_NAME = 'ai-edge-model-explorer'
MODULE_NAME = 'model_explorer'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8080
DEFAULT_COLAB_HEIGHT = 850
DEFAULT_SETTINGS = {
    'const_element_count_limit': 16,
}



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Type, TypedDict
from typing_extensions import NotRequired

from .extension_matadata import ExtensionMetadata
from .graph_builder import Graph, GraphCollection


class ClassWithMetadata:
  metadata: ExtensionMetadata


ExtensionClassType = Type[ClassWithMetadata]

AdapterCommand = TypedDict(
    'AdapterCommand',
    {
        'cmdId': str,
        'modelPath': str,
        'settings': Dict,
        'deleteAfterConversion': bool,
    },
)

ModelExplorerGraphs = TypedDict(
    'ModelExplorerGraphs',
    {
        'graphs': NotRequired[list[Graph]],
        'graphCollections': NotRequired[list[GraphCollection]],
    },
)



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import os
from typing import Dict, TypedDict, Union
from urllib.parse import quote

import requests
import torch
from typing_extensions import NotRequired

from .consts import DEFAULT_HOST, DEFAULT_SETTINGS
from .node_data_builder import NodeData
from .pytorch_exported_program_adater_impl import PytorchExportedProgramAdapterImpl
from .types import ModelExplorerGraphs

ModelSource = TypedDict(
    'ModelSource', {'url': str, 'adapterId': NotRequired[str]}
)

EncodedUrlData = TypedDict(
    'EncodedUrlData',
    {
        'models': list[ModelSource],
        'nodeData': NotRequired[list[str]],
        'nodeDataTargets': NotRequired[list[str]],
    },
)


class ModelExplorerConfig:
  """Stores the data to be visualized in Model Explorer."""

  def __init__(self) -> None:
    self.model_sources: list[ModelSource] = []
    self.graphs_list: list[ModelExplorerGraphs] = []
    self.node_data_sources: list[str] = []
    # List of model names to apply node data to. For the meaning of
    # "model name", see comments in `add_node_data_from_path` method below.
    self.node_data_target_models: list[str] = []
    self.node_data_list: list[Union[NodeData, str]] = []
    self.reuse_server_host = ''
    self.reuse_server_port = -1

  def add_model_from_path(
      self, path: str, adapterId: str = ''
  ) -> 'ModelExplorerConfig':
    """Adds a model path to the config.

    Args:
      path: the model path to add.
      adapterId: the id of the adapter to use. Default is empty string meaning
        it will use the default adapter.
    """
    # Get the absolute path (after expanding home dir path "~").
    abs_model_path = os.path.abspath(os.path.expanduser(path))

    # Construct model source and add it.
    model_source: ModelSource = {'url': abs_model_path}
    if adapterId != '':
      model_source['adapterId'] = adapterId
    self.model_sources.append(model_source)

    return self

  def add_model_from_pytorch(
      self,
      name: str,
      exported_program: torch.export.ExportedProgram,
      settings=DEFAULT_SETTINGS,
  ) -> 'ModelExplorerConfig':
    """Adds the given pytorch model.

    Args:
      name: the name of the model for display purpose.
      exported_program: the ExportedProgram from torch.export.export.
      settings: The settings that config the visualization.
    """
    # Convert the given model to model explorer graphs.
    print('Converting pytorch model to model explorer graphs...')
    adapter = PytorchExportedProgramAdapterImpl(exported_program, settings)
    graphs = adapter.convert()
    graphs_index = len(self.graphs_list)
    self.graphs_list.append(graphs)

    # Construct model source.
    #
    # The model source has a special format, in the form of:
    # graphs://{name}/{graphs_index}
    model_source: ModelSource = {'url': f'graphs://{name}/{graphs_index}'}
    self.model_sources.append(model_source)

    return self

  def add_node_data_from_path(
      self, path: str, model_name: Union[str, None] = None
  ) -> 'ModelExplorerConfig':
    """Adds node data file to the config.

    Args:
      path: the path of the node data json file to add.
      model_name: the name of the model. If not set, the node data will be
          applied to the first model added to the config by default.

          To set this parameter:
          For non-pytorch model, this should be the name of the model file
          (e.g. model.tflite). For pytorch model, it should be the `name`
          parameter used to call the `add_model_from_pytorch` api.
    """
    # Get the absolute path (after expanding home dir path "~").
    abs_model_path = os.path.abspath(os.path.expanduser(path))

    self.node_data_sources.append(abs_model_path)
    if model_name is None:
      self.node_data_target_models.append('')
    else:
      self.node_data_target_models.append(model_name)

    return self

  def add_node_data(
      self,
      name: str,
      node_data: Union[NodeData, str],
      model_name: Union[str, None] = None,
  ) -> 'ModelExplorerConfig':
    """Adds the given node data object.

    Args:
      name: the name of the NodeData for display purpose.
      node_data: the NodeData object or node data json string to add.
      model_name: the name of the model. If not set, the node data will be
          applied to the first model added to the config by default.

          To set this parameter:
          For non-pytorch model, this should be the name of the model file
          (e.g. model.tflite). For pytorch model, it should be the `name`
          parameter used to call the `add_model_from_pytorch` api.
    """
    node_data_index = len(self.node_data_list)
    self.node_data_list.append(node_data)

    # Construct sources.
    #
    # The node data source has a special format, in the form of:
    # node_data://{name}//{index}
    self.node_data_sources.append(f'node_data://{name}/{node_data_index}')
    if model_name is None:
      self.node_data_target_models.append('')
    else:
      self.node_data_target_models.append(model_name)
    return self

  def set_reuse_server(
      self,
      server_host: str = DEFAULT_HOST,
      server_port: Union[int, None] = None,
  ):
    """Makes it to reuse the existing server instead of starting a new one.

    Args:
      server_host: the host of the server to reuse.
      server_port: the port of the server to reuse. If unspecified, it will try
          to find a running server from port 8080 to 8099.
    """
    # Find the server to reuse.
    self.reuse_server_host = server_host
    if server_port is None or server_port < 0:
      self.reuse_server_port = self._find_running_server_port(host=server_host)
    else:
      if self._check_running_server(host=server_host, port=server_port):
        self.reuse_server_port = server_port

    if self.reuse_server_port > 0:
      print(
          'Re-using running server at'
          f' http://{self.reuse_server_host}:{self.reuse_server_port}'
      )
    else:
      print(f'No running server found. Will start a new server.')

    return self

  def to_url_param_value(self) -> str:
    """Converts the config to the url param value."""
    # Construct url data.
    encoded_url_data: EncodedUrlData = {'models': self.model_sources}

    if self.node_data_sources:
      encoded_url_data['nodeData'] = self.node_data_sources
    if self.node_data_target_models:
      encoded_url_data['nodeDataTargets'] = self.node_data_target_models

    # Return its json string.
    return quote(json.dumps(encoded_url_data))

  def get_model_explorer_graphs(self, index: int) -> ModelExplorerGraphs:
    return self.graphs_list[index]

  def get_node_data(self, index: int) -> Union[NodeData, str]:
    return self.node_data_list[index]

  def has_data_to_encode_in_url(self) -> bool:
    return len(self.model_sources) > 0 or len(self.node_data_sources) > 0

  def get_transferrable_data(self) -> Dict:
    return {
        'model_sources': self.model_sources,
        'node_data_sources': self.node_data_sources,
        'node_data_target_models': self.node_data_target_models,
    }

  def set_transferrable_data(self, data: Dict):
    self.model_sources = data['model_sources']
    self.node_data_sources = data['node_data_sources']
    self.node_data_target_models = data['node_data_target_models']

  def _find_running_server_port(self, host: str = DEFAULT_HOST) -> int:
    for port in range(8080, 8100):
      if self._check_running_server(host=host, port=port):
        return port
    return -1

  def _check_running_server(self, host: str, port: int) -> bool:
    try:
      resp = requests.get(f'http://{host}:{port}/check_health')
      return resp.status_code == 200 and resp.text == 'model_explorer_ok'
    except:
      return False



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import TypedDict, Union

import torch
from typing_extensions import NotRequired

from . import server
from .config import ModelExplorerConfig, NodeData
from .consts import (
    DEFAULT_COLAB_HEIGHT,
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_SETTINGS,
)

NodeDataInfo = TypedDict(
    'NodeDataInfo',
    {
        # The name of the node data for display purpose.
        'name': str,
        # The NodeData object of node data json string to add.
        #
        # This field takes precedence over node_data_path field below when they
        # are both set.
        'node_data': NotRequired[Union[NodeData, str]],
        # The path of the node data json file to add.
        'node_data_path': NotRequired[str],
        # The name of the model to apply the node data to. If not set, the node
        # data will be applied to the first model by default.
        #
        # To set this field, use the name of the model file (e.g. model.tflite).
        'model_name': NotRequired[str],
    },
)


def config() -> ModelExplorerConfig:
  """Create a new config object."""
  return ModelExplorerConfig()


def visualize(
    model_paths: Union[str, list[str]] = [],
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    extensions: list[str] = [],
    node_data: Union[NodeDataInfo, list[NodeDataInfo]] = [],
    colab_height=DEFAULT_COLAB_HEIGHT,
    reuse_server: bool = False,
    reuse_server_host: str = DEFAULT_HOST,
    reuse_server_port: Union[int, None] = None,
) -> None:
  """Starts the ME local server and visualizes the models by the given paths.

  Args:
    model_paths: A model path or a list of model paths to visualize.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    extensions: List of extension names to be run with model explorer.
    node_data: The node data or a list of node data to display.
    colab_height: The height of the embedded iFrame when running in colab.
    reuse_server: Whether to reuse the current server/browser tab(s) to
        visualize.
    reuse_server_host: the host of the server to reuse. Default to localhost.
    reuse_server_port: the port of the server to reuse. If unspecified, it will
        try to find a running server from port 8080 to 8099.
  """
  # Construct config.
  cur_config = config()
  model_paths_list = model_paths

  if isinstance(model_paths, str):
    model_paths_list = [model_paths]
  for model_path in model_paths_list:
    cur_config.add_model_from_path(path=model_path)

  _add_node_data_to_config(node_data=node_data, config=cur_config)

  if reuse_server:
    cur_config.set_reuse_server(
        server_host=reuse_server_host, server_port=reuse_server_port
    )

  # Start server.
  server.start(
      host=host,
      port=port,
      config=cur_config,
      colab_height=colab_height,
      extensions=extensions,
  )


def visualize_pytorch(
    name: str,
    exported_program: torch.export.ExportedProgram,
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    extensions: list[str] = [],
    node_data: Union[NodeDataInfo, list[NodeDataInfo]] = [],
    colab_height=DEFAULT_COLAB_HEIGHT,
    settings=DEFAULT_SETTINGS,
) -> None:
  """Visualizes a pytorch model.

  Args:
    name: The name of the model for display purpose.
    exported_program: The ExportedProgram from torch.export.export.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    extensions: List of extension names to be run with model explorer.
    node_data: The node data or a list of node data to display.
    colab_height: The height of the embedded iFrame when running in colab.
    settings: The settings that config the visualization.
  """
  # Construct config.
  cur_config = config()
  cur_config.add_model_from_pytorch(
      name, exported_program=exported_program, settings=settings
  )

  _add_node_data_to_config(node_data=node_data, config=cur_config)

  # Start server.
  server.start(
      host=host,
      port=port,
      config=cur_config,
      colab_height=colab_height,
      extensions=extensions,
  )


def visualize_from_config(
    config: Union[ModelExplorerConfig, None] = None,
    host=DEFAULT_HOST,
    port=DEFAULT_PORT,
    extensions: list[str] = [],
    cors_host: Union[str, None] = None,
    no_open_in_browser: bool = False,
    colab_height=DEFAULT_COLAB_HEIGHT,
) -> None:
  """Visualizes with a config.

  Args:
    config: the object that stores the models to be visualized.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    extensions: List of extension names to be run with model explorer.
    cors_host: The value of the Access-Control-Allow-Origin header. The header
      won't be present if it is None.
    no_open_in_browser: Don't open the web app in browser after server starts.
    colab_height: The height of the embedded iFrame when running in colab.
  """
  # Start server.
  server.start(
      host=host,
      port=port,
      config=config,
      extensions=extensions,
      cors_host=cors_host,
      no_open_in_browser=no_open_in_browser,
      colab_height=colab_height,
  )


def _add_node_data_to_config(
    node_data: Union[NodeDataInfo, list[NodeDataInfo]],
    config: ModelExplorerConfig,
):
  # Convert NodeDataInfo to [NodeDataInfo] if necessary.
  node_data_list: list[NodeDataInfo] = []
  if isinstance(node_data, list):
    node_data_list = node_data
  else:
    node_data_list = [node_data]

  for node_data_info in node_data_list:
    name = node_data_info.get('name', 'node data')
    node_data_path = node_data_info.get('node_data_path')
    node_data_obj = node_data_info.get('node_data')
    model_name = node_data_info.get('model_name')
    if node_data_obj:
      config.add_node_data(
          name=name, node_data=node_data_obj, model_name=model_name
      )
    elif node_data_path:
      config.add_node_data_from_path(path=node_data_path, model_name=model_name)
